--
-- Magnitude55
--
-- @author:    	Xentro (www.fs-uk.com)(Marcus@Xentro.se)
-- @version:    0.95
-- @date:       2013-10-12
-- 

Magnitude55 = {};

Magnitude55.ATTACH_MODE = 0;
Magnitude55.LIFTED_MODE = 1;
Magnitude55.LOCKED_MODE = 2;
Magnitude55.PENDING_MODE = 3;

Magnitude55.LOCKED = 0;
Magnitude55.GOOSENECK_UNLOCK = 1;
Magnitude55.BED_UNLOCK = 2;

function Magnitude55.initSpecialization()
	Vehicle.registerJointType("extraAxle");
end

function Magnitude55.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Cylindered, specializations) and SpecializationUtil.hasSpecialization(AnimatedVehicle, specializations);
end;

function Magnitude55:load(xmlFile)
	self.changeStep = SpecializationUtil.callSpecializationsFunction("changeStep");
	self.changeUnlockMode = SpecializationUtil.callSpecializationsFunction("changeUnlockMode");
	self.setGoosAnimation = SpecializationUtil.callSpecializationsFunction("setGoosAnimation");
	self.setRideHeight = SpecializationUtil.callSpecializationsFunction("setRideHeight");
	self.setNewRotLimit = SpecializationUtil.callSpecializationsFunction("setNewRotLimit");
	self.rampCallback = Magnitude55.rampCallback;
	self.bedIsAttached = Magnitude55.bedIsAttached;
	self.bedInteractionCallback = Magnitude55.bedInteractionCallback;
	self.setLightsTypesMask = Utils.overwrittenFunction(self.setLightsTypesMask, Magnitude55.newSetLightsTypesMask);
	self.setBrakeLightsVisibility = Utils.overwrittenFunction(self.setBrakeLightsVisibility, Magnitude55.newSetBrakeLightsVisibility);
	
	self.inputButtons = {InputBinding.IMPLEMENT_EXTRA, InputBinding.IMPLEMENT_EXTRA2, InputBinding.CAMERA_ZOOM_OUT, InputBinding.CAMERA_ZOOM_IN};
	
	self.gooseneckMovingToolId = 1;
	self.modeLimits = {0.7, 0.9}; -- lifted, locked (in percentage)
	self.currentMode = Magnitude55.LOCKED_MODE;
	self.currentUnlockMode = Magnitude55.GOOSENECK_UNLOCK;
	
	assert(self.movingTools[self.gooseneckMovingToolId] ~= nil, "There are no Moving tool defined for self.gooseneckMovingToolId");
	
	self.cargoAreaAnimationId = 1;
	
	-- attach and lifting/lowering Magnitude55 - customJoints.lua
	self.customJointSetting = {1, 1, 2, 5.2, {1.0, -0.2}}; -- parentJointId, childJointId, jointAxis, attachWindow, (attachWindowOffset, attachWindowOffset)
	self.jointLimits = {0, 0, 0, 20} -- rotLimits -locked, locked, -unlocked, unlocked
	self.lastAttachValues = {{false, 0}, {false, 0}, {false, 0}, 0};
	self.lastRot = 0;
	
	self.hoses = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.gooseneck#hoses"));
	if self.hoses ~= nil then
		setVisibility(self.hoses, false);
	end;
	
	self.hatch = {};
	self.hatch.node = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.gooseneck#hatch"));
	self.hatch.open = {0, 0, 0};
	self.hatch.closed = {0, 0, math.rad(-73)};
	if self.hatch.node ~= nil then
		setRotation(self.hatch.node, unpack(self.hatch.closed));
	end;
	
	self.liftArm = {name = getXMLString(xmlFile, "vehicle.gooseneck#liftArmAnimation"), bool = false};
	
	self.rideHeights = {};
	self.currentRideHeight = 0;
	local i = 0;
	while true do
		local key = string.format("vehicle.gooseneck.rideHeight(%d)", i);
		if not hasXMLProperty(xmlFile, key) then break; end;
		
		local name = getXMLString(xmlFile, key .. "#animationName");
		if name ~= nil and self.animations[name] ~= nil then
			local entry = {}
			entry.animationName = name;
			entry.rotLimit = math.rad(getXMLFloat(xmlFile, key .. "#rotLimit"));
			
			if Utils.getNoNil(getXMLBool(xmlFile, key .. "#start"), false) then
				self.currentRideHeight = i + 1;
			end;
			
			table.insert(self.rideHeights, entry);
			i = i + 1;
		else
			break;
		end;
	end;
	
	self.rideSettings = {};
	self.rideSettings.max = self.movingTools[self.gooseneckMovingToolId].rotMax;
	self.rideSettings.heights = table.getn(self.rideHeights);
	
	assert(self.rideSettings.heights ~= 0, "There are no rideHeights defined!");
	
	self.ramps = {};
	local i = 0;
	while true do
		local key = string.format("vehicle.bed.ramp(%d)", i);
		if not hasXMLProperty(xmlFile, key) then break; end;
		
		local name = getXMLString(xmlFile, key .. "#animationName");
		if name ~= nil and self.animations[name] ~= nil then
			local entry = {}
			entry.animationName = name;
			entry.contactNode = Utils.indexToObject(self.components, getXMLString(xmlFile, key .. "#contactNode"));
			if self.isServer then
				addContactReport(entry.contactNode, 0.0001, "rampCallback", self);
			end;
			
			table.insert(self.ramps, entry);
			i = i + 1;
		else
			break;
		end;
	end;
	
	if self.interactionLocations ~= nil then
		for node in pairs(self.interactionLocations) do
			self.playerInteractionTrigger = node;
			break;
		end;
	end;
	
	self.playerInBedRange = false;
	self.playerBedInteractionTrigger = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.bed.interaction#trigger"));
	if self.playerBedInteractionTrigger ~= nil then
		addTrigger(self.playerBedInteractionTrigger, "bedInteractionCallback", self);
	end;
		
	if self.isRealistic ~= nil and self.isRealistic then
		RealisticVehicle.registerExtraMassGetter(self, Magnitude55.getCurrentMass);
		self.totalMassLoad = 0;
		self.totalMassObjects = {};
	end;
end;

function Magnitude55:delete()
	if self.isServer then
		for _, v in ipairs(self.ramps) do
			removeContactReport(v.contactNode);
		end;
	end;
	if self.playerBedInteractionTrigger ~= nil then
		removeTrigger(self.playerBedInteractionTrigger);
	end;
	self.playerInteractionTrigger = nil;
end;

function Magnitude55:readStream(streamId, connection)
	self:changeUnlockMode(streamReadUIntN(streamId, 2), true);
	self:setGoosAnimation(streamReadBool(streamId), true);
	self:setRideHeight(streamReadInt8(streanId), true);
end;

function Magnitude55:writeStream(streamId, connection)
	streamWriteUIntN(streamId, self.currentUnlockMode, 2);
	streamWriteBool(streamId, self.liftArm.bool);
	streamWriteInt8(streamId, self.currentRideHeight);
end;

function Magnitude55:mouseEvent(posX, posY, isDown, isUp, button)
end;

function Magnitude55:keyEvent(unicode, sym, modifier, isDown)
end;

function Magnitude55:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)
	if not resetVehicles then
		local currentUnlockMode = Utils.getNoNil(getXMLInt(xmlFile, key .. "#currentUnlockMode"), Magnitude55.GOOSENECK_UNLOCK);
		self:changeUnlockMode(currentUnlockMode, true);
		
		local currentRideHeight = Utils.getNoNil(getXMLInt(xmlFile, key .. "#currentRideHeight"), self.currentRideHeight);
		self:setRideHeight(currentRideHeight, true, true);
		
		local liftArm = getXMLBool(xmlFile, key .. "#liftArm");
		local liftArmTime = getXMLFloat(xmlFile, key .. "#liftArmTime");
		if liftArm ~= nil and liftArmTime ~= nil then
			local speed = 1;
			if liftArmTime < 0.5 then
				speed = -1;
			end
			
			self.liftArm.bool = liftArm;
			self:playAnimation(self.liftArm.name, speed);
			self:setAnimationTime(self.liftArm.name, liftArmTime);
			self:updateCylinderedInitial();
		end;
		
		for i = 1, table.getn(self.ramps) do
			local rampKey = key .. string.format(".ramp(%d)", i - 1);
			if not hasXMLProperty(xmlFile, rampKey) then break; end;
			
			local animationName = getXMLString(xmlFile, rampKey .. "#animation");
			local animationTime = getXMLFloat(xmlFile, rampKey .. "#animationTime");
			
			if animationName ~= nil then
				local speed = 1;
				if animationTime < 0.5 then
					speed = -1;
				end
				
				self:playAnimation(animationName, speed);
				self:setAnimationTime(animationName, animationTime);
				self:updateCylinderedInitial();
			end;
		end;
		
		if self.isRealistic ~= nil and self.isRealistic then
			local mass = getXMLFloat(xmlFile, key .. "#totalMassLoad");
			
			if mass ~= nil then
				self.totalMassLoad = mass;
			end;
		end;
	end; 

    return BaseMission.VEHICLE_LOAD_OK;
end;

function Magnitude55:getSaveAttributesAndNodes(nodeIdent)	
	local attributes = 'currentUnlockMode="' .. self.currentUnlockMode .. '" currentRideHeight="' .. self.currentRideHeight .. '" liftArm="' .. tostring(self.liftArm.bool) .. '" liftArmTime="' .. self:getAnimationTime(self.liftArm.name) .. '"';
	local nodes = "";
	
	if table.getn(self.ramps) > 0 then
		for i, v in ipairs(self.ramps) do
			if i > 1 then
				nodes = nodes .. "\n";
			end;
			
			nodes = nodes .. nodeIdent .. '<ramp';
			nodes = nodes .. ' animation="' .. v.animationName .. '"';
			nodes = nodes .. ' animationTime="' .. self:getAnimationTime(v.animationName) .. '"';
			nodes = nodes .. " />"
		end;
	end;
	
	if self.isRealistic ~= nil and self.isRealistic then
		attributes = attributes .. 'totalMassLoad="' .. self.totalMassLoad .. '"';
	end;
	
    return attributes, nodes;
end;

function Magnitude55:update(dt)
	if self.isClient then
		local playerControlled = false;
		
		if self.interactionLocations ~= nil and self.playerInteractionTrigger ~= nil then
			if g_gui.currentGui == nil and not g_currentMission.isPlayerFrozen then
				playerControlled = self.interactionLocations[self.playerInteractionTrigger].playerCanControl;
			end;
		end;
		
		if (playerControlled and g_currentMission.controlPlayer) or (self:getIsActive() and self:getIsActiveForInput(true)) then
			if self:bedIsAttached() then
				if self.currentMode == Magnitude55.LIFTED_MODE then
					if self.currentUnlockMode == Magnitude55.LOCKED then
						if InputBinding.hasEvent(self.inputButtons[1]) then
							self:changeUnlockMode(Magnitude55.GOOSENECK_UNLOCK);
						elseif InputBinding.hasEvent(self.inputButtons[2]) then
							if self.attacherVehicle ~= nil then
								self:changeUnlockMode(Magnitude55.BED_UNLOCK);
							end;
						end;
						
						if g_i18n:hasText("PREPARE_DEATCH_1") and g_i18n:hasText("PREPARE_DEATCH_2") then 
							g_currentMission:addHelpButtonText(g_i18n:getText("PREPARE_DEATCH_1"), self.inputButtons[1]);
							if self.attacherVehicle ~= nil then
								g_currentMission:addHelpButtonText(g_i18n:getText("PREPARE_DEATCH_2"), self.inputButtons[2]);
							end;
						end;
					else
						if InputBinding.hasEvent(self.inputButtons[1]) or InputBinding.hasEvent(self.inputButtons[2]) then
							self:changeUnlockMode(Magnitude55.LOCKED);
						end;
						
						if g_i18n:hasText("LOCK_RIDE_HEIGHT") then
							g_currentMission:addHelpButtonText(g_i18n:getText("LOCK_RIDE_HEIGHT"), self.inputButtons[2]);
						end;
						
						if InputBinding.hasEvent(self.inputButtons[4]) then
							self:setRideHeight(self.currentRideHeight + 1);
						elseif InputBinding.hasEvent(self.inputButtons[3]) then
							self:setRideHeight(self.currentRideHeight - 1);
						end;
						if g_i18n:hasText("CURRENT_RIDE_HEIGHT") and g_i18n:hasText("ADJUST_RIDE_HEIGHT") then
							g_currentMission:addExtraPrintText(g_i18n:getText("ADJUST_RIDE_HEIGHT"));
							g_currentMission:addExtraPrintText(string.format(g_i18n:getText("CURRENT_RIDE_HEIGHT"), self.currentRideHeight));
						end;
					end;
				end;
			else
				if InputBinding.hasEvent(self.inputButtons[2]) then
					self:setGoosAnimation(not self.liftArm.bool);
				end;
				
				if g_i18n:hasText("TRUCK_SUPPORT") then
					g_currentMission:addHelpButtonText(g_i18n:getText("TRUCK_SUPPORT"), self.inputButtons[2]);
				end;
			end;
		end;
		
		if self.playerInBedRange and g_currentMission.controlPlayer then 
			if g_gui.currentGui == nil and not g_currentMission.isPlayerFrozen then
				if self.setIsBaleAttached ~= nil then
					if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA3) then
						self:setIsBaleAttached(not self.isBaleAttach);
					end;
					
					if not self.isBaleAttach then
						if g_i18n:hasText("Unlock") then 
							g_currentMission:addHelpButtonText(g_i18n:getText("Unlock"), InputBinding.IMPLEMENT_EXTRA3);
						end;
					else
						if g_i18n:hasText("Lock") then 
							g_currentMission:addHelpButtonText(g_i18n:getText("Lock"), InputBinding.IMPLEMENT_EXTRA3);
						end;
					end;
				end;
				
				if self.clipAnimations ~= nil then
					local id = self.cargoAreaAnimationId;
					local anim = self.clipAnimations[id];
					
					if anim ~= nil then
						if InputBinding.hasEvent(anim.inputBinding) then
							self:setAnimationClip(id, not anim.isActivated);
						end;
						
						if anim.text[1] ~= nil then
							g_currentMission:addHelpButtonText(anim.text[1], anim.inputBinding);
						elseif anim.text[2] ~= nil then
							if anim.isActivated then
								g_currentMission:addHelpButtonText(anim.text[3], anim.inputBinding);
							else
								g_currentMission:addHelpButtonText(anim.text[2], anim.inputBinding);
							end;
						end;
					end;
				end;
			end;
		end;
	end;
end;

function Magnitude55:updateTick(dt)
	local tool = self.movingTools[self.gooseneckMovingToolId];
	if tool ~= nil then
		local rot = tool.curRot[tool.rotationAxis];
		
		if self.isServer then
			if rot ~= self.lastRot then
				if self.updateJoint ~= nil then
					self:updateJoint(self.customJointSetting[1]);
				end;
				
				self.lastRot = rot;
			end;
		end;
		
		if rot <= (tool.rotMin * self.modeLimits[1]) then -- LIFTED_MODE
			if self.currentMode ~= Magnitude55.LIFTED_MODE then
				self:changeStep(Magnitude55.LIFTED_MODE);
			end;
			
			self:setNewRotLimit(false);
		else
			if self.currentUnlockMode == Magnitude55.LOCKED then
				if rot >= (tool.rotMax * self.modeLimits[2]) then -- LOCKED_MODE
					if self.currentMode ~= Magnitude55.LOCKED_MODE then
						self:changeStep(Magnitude55.LOCKED_MODE);
					end;
				else -- PENDING_MODE
					if self.currentMode ~= Magnitude55.PENDING_MODE then
						self:changeStep(Magnitude55.PENDING_MODE);
					end;
				end;
			else -- ATTACH_MODE
				if self.currentMode ~= Magnitude55.ATTACH_MODE then
					self:changeStep(Magnitude55.ATTACH_MODE);
				end;
			end;
		end;
	end;
	
	if self.parentJoints ~= nil and self.parentJoints[self.customJointSetting[1]] ~= nil then
		local rotLimitCorrect = true;
		
		if not self:bedIsAttached() then
			local limit = 0 + self.customJointSetting[4] - self.customJointSetting[5][1];
			local limit2 = 360 - self.customJointSetting[4] - self.customJointSetting[5][2];
			
			local x_rot ,y_rot, z_rot = getWorldRotation(self.childJoints[self.customJointSetting[2]].node);
			local x_rot2 ,y_rot2, z_rot2 = getWorldRotation(self.parentJoints[self.customJointSetting[1]].node);
			
			local xB, x = Magnitude55.countLimit(x_rot - x_rot2, limit, limit2);
			local yB, y = Magnitude55.countLimit(y_rot - y_rot2, limit, limit2);
			local zB, z = Magnitude55.countLimit(z_rot - z_rot2, limit, limit2);
			
			rotLimitCorrect = xB and yB and zB;
			local x2, y2, z2 = getWorldTranslation(self.childJoints[self.customJointSetting[2]].node);
			local x3, y3, z3 = getWorldTranslation(self.parentJoints[self.customJointSetting[1]].node);
			self.lastAttachValues = {{xB, x}, {yB, y}, {zB, z}, Utils.vector3Length(x2 - x3, y2 - y3, z2 - z3)};
		else
			if self.liftArm.bool then
				self:setGoosAnimation(false, true);
			end;
		end;
		
		if not self:bedIsAttached() and not rotLimitCorrect then
			self.parentJoints[self.customJointSetting[1]].allow[1] = false;
			if g_i18n:hasText("ATTACH_ERROR") then
				self.parentJoints[self.customJointSetting[1]].error = g_i18n:getText("ATTACH_ERROR");
			end;
		elseif self.currentMode == Magnitude55.ATTACH_MODE then
			if self.currentUnlockMode == Magnitude55.BED_UNLOCK then
				self.parentJoints[self.customJointSetting[1]].allow[1] = true;
			else
				self.parentJoints[self.customJointSetting[1]].allow[1] = false;
				if g_i18n:hasText("ATTACH_ERROR_2") then
					self.parentJoints[self.customJointSetting[1]].error = g_i18n:getText("ATTACH_ERROR_2");
				end;
			end;
		else
			self.parentJoints[self.customJointSetting[1]].allow[1] = false;
			if g_i18n:hasText("UNHOOK_ERROR") then
				self.parentJoints[self.customJointSetting[1]].error = g_i18n:getText("UNHOOK_ERROR");
			end;
		end;
	end;
	
	if self.isServer then
		if self.levers ~= nil and self.levers[2] ~= nil then
			if self.liftArm.name ~= nil then
				if self:getIsAnimationPlaying(self.liftArm.name) then
					local name = self.animations[self.liftArm.name];
					
					if name ~= nil then
						if self.levers[2].moveDir ~= name.currentSpeed then
							self.levers[2].moveDir = name.currentSpeed;
							self:raiseDirtyFlags(self.moveDirDirtyFlag);
						end;
					end;
				else
					if self.levers[2].moveDir ~= 0 then
						self.levers[2].moveDir = 0;
						self:raiseDirtyFlags(self.moveDirDirtyFlag);
					end;
				end;
			end;
		end;
		
		if not self:bedIsAttached() then
			if not self.isBraking then
				self:onBrake(true);
			end;
			
			if self.rLights ~= nil then
				if self.rLights.bll.a or self.rLights.blr.a then
					self:setBlinkerLState(false);
					self:setBlinkerRState(false);
				end;
			end;
		end;
	end;
end;

function Magnitude55:draw()
	if not self:bedIsAttached() then
		if self.lastAttachValues[4] < 1.75 then
			if self.lastAttachValues[1][1] and self.lastAttachValues[2][1] and self.lastAttachValues[3][1] then
				g_currentMission:enableHudIcon("attach", 3);
			else
				if not self.lastAttachValues[1][1] then
					if g_i18n:hasText("ERROR_LEANS") then
						g_currentMission:addExtraPrintText(string.format(g_i18n:getText("ERROR_LEANS"), self.lastAttachValues[1][2]));
					end;
				end;
				if not self.lastAttachValues[2][1] then
					if g_i18n:hasText("ERROR_LINE") then
						g_currentMission:addExtraPrintText(string.format(g_i18n:getText("ERROR_LINE"), self.lastAttachValues[2][2]));
					end;
				end;
				if not self.lastAttachValues[3][1] then
					if g_i18n:hasText("ERROR_ROTATION") then
						g_currentMission:addExtraPrintText(string.format(g_i18n:getText("ERROR_ROTATION"), self.lastAttachValues[3][2]));
					end;
				end;
			end;
		end;
	end;
end;

function Magnitude55:changeStep(value)
	self.currentMode = value;
end;

function Magnitude55:changeUnlockMode(id, noEventSend)
	self.currentUnlockMode = id;
	local animation = self.rideHeights[self.currentRideHeight].animationName
	
	if id == Magnitude55.LOCKED then
		self:playAnimation(animation, 1, nil, true)
	else
		self:playAnimation(animation, -1, nil, true)
	end;
	
	changeUnlockModeEvent.sendEvent(self, id, noEventSend);
end;

function Magnitude55:setGoosAnimation(bool, noEventSend)
	self.liftArm.bool = bool;
	
	if bool then
		self:playAnimation(self.liftArm.name, 1, nil, true)
	else
		self:playAnimation(self.liftArm.name, -1, nil, true)
	end;
	
	setGoosAnimationEvent.sendEvent(self, bool, noEventSend);
end;

function Magnitude55:setRideHeight(id, noEventSend, forceLimit)
	if id == nil or id <= 0 then
		id = 1;
	elseif id > self.rideSettings.heights then
		id = self.rideSettings.heights;
	end;
	
	if self.currentRideHeight ~= id or forceLimit ~= nil and forceLimit then
		self.currentRideHeight = id;
		
		if forceLimit ~= nil and forceLimit then
			self:setNewRotLimit(true);
		end;
		
		setRideHeightEvent.sendEvent(self, bool, forceLimit, noEventSend);
	end;
end;

function Magnitude55:setNewRotLimit(force)
	local tool = self.movingTools[self.gooseneckMovingToolId];
	local newRot = self.rideHeights[self.currentRideHeight].rotLimit;
	
	if self.currentUnlockMode ~= Magnitude55.LOCKED then
		newRot = self.rideSettings.max;
	end;
	
	if tool.rotMax ~= newRot or force then
		tool.rotMax = newRot;
		
		-- if self.isServer then
			if self:bedIsAttached() then
				if self.currentUnlockMode ~= Magnitude55.BED_UNLOCK then
					setJointRotationLimit(self.usedJoints[self.customJointSetting[1]].jointIndex, self.customJointSetting[3], true, Utils.degToRad(self.jointLimits[1]), Utils.degToRad(self.jointLimits[2]));
					print("~= bed unlocked");
				else
					setJointRotationLimit(self.usedJoints[self.customJointSetting[1]].jointIndex, self.customJointSetting[3], true, Utils.degToRad(self.jointLimits[3]), Utils.degToRad(self.jointLimits[4]));
					print("== bed unlocked");
				end;
			end;
		-- end;
	end;
end;

function Magnitude55:rampCallback(objectId, otherObjectId, isStart, normalForce, tangentialForce)
	if self.isServer then
		for _, v in ipairs(self.ramps) do
			if v.contactNode == objectId then
				if ((isStart or normalForce > 0.001 or tangentialForce > 0.001) and otherObjectId ~= self.components[4]) then
					if g_currentMission.player ~= nil and otherObjectId ~= g_currentMission.player.rootNode then
						if self:getIsAnimationPlaying(v.animationName) then
							self:stopAnimation(v.animationName);
						end;
					end;
				else
					if not self:bedIsAttached() then
						if self:getAnimationTime(v.animationName) ~= 1 then
							if not self:getIsAnimationPlaying(v.animationName) then
								self:playAnimation(v.animationName, 1);
							end;
						end;
					end;
				end;
			end;
		end;
	end;
end;

function Magnitude55:bedIsAttached()
	if self.usedJoints ~= nil then
		return self.usedJoints[self.customJointSetting[1]] ~= nil and self.usedJoints[self.customJointSetting[1]].jointIndex ~= nil;
	end;
	
	return false;
end;

function Magnitude55:bedInteractionCallback(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
	if onEnter or onLeave then
		if g_currentMission.player ~= nil and otherId == g_currentMission.player.rootNode then
			if onEnter then
				self.playerInBedRange = true;
			else
				self.playerInBedRange = false;
			end;
		end;
		
		if self.isRealistic ~= nil and self.isRealistic then
			local object = g_currentMission.nodeToVehicle[otherId];
			if object == nil then
				object = g_currentMission:getNodeObject(otherId);
			end;
			
			if object ~= nil and object ~= self then
				local mass = 0;
				
				if object.isRealistic ~= nil and object.isRealistic then
					if object.realComponentsCurrentMass ~= nil then
						mass = object.realComponentsCurrentMass;
					elseif object.realCurrentMass ~= nil then
						mass = object.realCurrentMass;
					end;
				else
					-- get density
				end;
				
				if onEnter then
					if self.totalMassObjects[object] == nil then
						self.totalMassLoad = self.totalMassLoad + mass;
						self.totalMassObjects[object] = mass;
						
						-- print("totalMassLoad = " .. self.totalMassLoad .. ", + " .. mass .. " name " .. tostring(object.realVehicleName));
					end;
				else
					if self.totalMassObjects[object] ~= nil then
						self.totalMassLoad = math.max((self.totalMassLoad - mass), 0);
						self.totalMassObjects[object] = nil;
						
						-- print("totalMassLoad = " .. self.totalMassLoad .. ", - " .. mass .. " name " .. tostring(object.realVehicleName));
					end;
				end;
			end;
		end;
	end;
end;

function Magnitude55:newSetLightsTypesMask(superFunc, lightsTypesMask, noEventSend)
	if self:bedIsAttached() then
		superFunc(self, lightsTypesMask, noEventSend);
	else
		if self.lightsTypesMask ~= 0 then
			superFunc(self, 0, noEventSend);
		end;
	end;
end;

function Magnitude55:newSetBrakeLightsVisibility(superFunc, visibility)
	if self:bedIsAttached() then
		superFunc(self, visibility);
	else
		if self.brakeLightsVisibility ~= false then
			superFunc(self, false);
		end;
	end;
end;

function Magnitude55.countLimit(value, limit1, limit2)
	value = math.abs(math.deg(value));
	local bool = value > limit2 or value < limit1;
	
	return bool, math.floor(value);
end;

function Magnitude55.customJointAttach(self, jointId)
	self.isBraking = false;
	
	if self.hoses ~= nil then
		setVisibility(self.hoses, true);
	end;
	if self.hatch.node ~= nil then
		setRotation(self.hatch.node, unpack(self.hatch.open));
	end;
	for _, v in ipairs(self.ramps) do
		self:playAnimation(v.animationName, -1, nil, true);
	end;
	
	if self.isServer then
		if self.currentUnlockMode ~= Magnitude55.BED_UNLOCK then
			if jointId == self.customJointSetting[1] then
				setJointRotationLimit(self.usedJoints[self.customJointSetting[1]].jointIndex, self.customJointSetting[3], true, Utils.degToRad(self.jointLimits[1]), Utils.degToRad(self.jointLimits[2]));
			end;
		end;
	end;
	
	local rootVehicle = self:getRootAttacherVehicle();
	if rootVehicle ~= nil then
		if rootVehicle.lightsTypesMask ~= self.lightsTypesMask then
			self:setLightsTypesMask(rootVehicle.lightsTypesMask, true);
		end;
		
		if rootVehicle.brakeLightsVisibility ~= self.brakeLightsVisibility then
			self:setBrakeLightsVisibility(rootVehicle.brakeLightsVisibility);
		end;
	end;
	
	if self.clipAnimations ~= nil and self.clipAnimations[self.cargoAreaAnimationId] ~= nil then
		self.clipAnimations[self.cargoAreaAnimationId].canActivate = true;
	end;
	
	self.updateSteeringAxleAngle = true;
	self.activateAdvancedSteering = true;
end;

function Magnitude55.customJointDetach(self, jointId)
	if self.hoses ~= nil then
		setVisibility(self.hoses, false);
	end;
	if self.hatch.node ~= nil then
		setRotation(self.hatch.node, unpack(self.hatch.closed));
	end;
	for _, v in ipairs(self.ramps) do
		self:playAnimation(v.animationName, 1, nil, true);
	end;
	if self.lightsTypesMask ~= 0 then
		self:setLightsTypesMask(0, true);
	end;
	if self.brakeLightsVisibility then
		self:setBrakeLightsVisibility(false);
	end;
	if self.clipAnimations ~= nil and self.clipAnimations[self.cargoAreaAnimationId] ~= nil then
		self.clipAnimations[self.cargoAreaAnimationId].canActivate = false;
	end;
	
	self.updateSteeringAxleAngle = false;
	self.activateAdvancedSteering = false;
end;

function Magnitude55:getCurrentMass()
	if not self:bedIsAttached() then
		return 0;
	end;
	
	return self.totalMassLoad;
end;