--
-- customJoints
--
-- @author:    	Xentro (www.fs-uk.com)(Marcus@Xentro.se)
-- @version:    v0.92
-- @date:       2013-10-19
-- 
--
--[[

<customJoints>
	<parentJoint type="myJointType" index="0>12" componentId="1" attachChildIdOnLoad="1" />
	
	<childJoint type="myJointType" index="0>12" componentId="1" name="Test" />
</customJoints>

]]--
customJoints = {};

function customJoints.prerequisitesPresent(specializations)
    return true;
end;

function customJoints:load(xmlFile)
	self.makeJoint = SpecializationUtil.callSpecializationsFunction("makeJoint");
	self.deleteJoint = SpecializationUtil.callSpecializationsFunction("deleteJoint");
	self.updateJoint = SpecializationUtil.callSpecializationsFunction("updateJoint");
	self.jointInRange = customJoints.jointInRange;
	
	self.usedJoints = {};
	
	self.childJoints = {};
	local i = 0;
	while true do
		local key = string.format("vehicle.customJoints.childJoint(%d)", i);
		if not hasXMLProperty(xmlFile, key) then break; end;
		
		local node = Utils.indexToObject(self.components, getXMLString(xmlFile, key .. "#index"));
		local jointType = getXMLString(xmlFile, key .. "#type");
		if node ~= nil and jointType ~= nil then
			local componentId = getXMLInt(xmlFile, key .. "#componentId");
			if componentId ~= nil and self.components[componentId + 1] ~= nil then
				local entry = {};
				entry.node = node;
				entry.type = string.lower(jointType);
				entry.componentId = componentId + 1;
				entry.name = Utils.getNoNil(getXMLString(xmlFile, key .. "#name"), "");
				
				table.insert(self.childJoints, entry);
				i = i + 1;
			else
				print("Error: could't find component " .. componentId + 1 .. " in '" .. self.configFileName .. "'");
				break;
			end;
		end;
	end;
	
	self.parentJoints = {};
	local i = 0;
	while true do
		local key = string.format("vehicle.customJoints.parentJoint(%d)", i);
		if not hasXMLProperty(xmlFile, key) then break; end;
		
		local node = Utils.indexToObject(self.components, getXMLString(xmlFile, key .. "#index"));
		local str = Utils.indexToObject(self.components, getXMLString(xmlFile, key .. "#str"));
		local jointType = getXMLString(xmlFile, key .. "#type");
		if node ~= nil and jointType ~= nil then
			local componentId = getXMLInt(xmlFile, key .. "#componentId");
			if componentId ~= nil and self.components[componentId + 1] ~= nil then
				local entry = {};
				entry.str = str;
				entry.node = node;
				entry.type = string.lower(jointType);
				entry.componentId = componentId + 1;
				entry.allow = {};
				entry.allow[1] = true;
				
				local attachChildIdOnLoad = getXMLInt(xmlFile, key .. "#attachChildIdOnLoad");
				if attachChildIdOnLoad ~= nil then
					if self.childJoints[attachChildIdOnLoad] ~= nil then
						if self.childJoints[attachChildIdOnLoad].type == entry.type then
							entry.attachChildIdOnLoad = attachChildIdOnLoad;
						else
							print("Error: attachChildIdOnLoad (" .. attachChildIdOnLoad .. ") don't have the same type as parent in '" .. self.configFileName .. "'");
							break;
						end;
					else
						print("Error: attachChildIdOnLoad (" .. attachChildIdOnLoad .. ") can't be found in '" .. self.configFileName .. "'");
						break;
					end;
				end;
				
				local rotLimit = Utils.getVectorNFromString(getXMLString(xmlFile, key .. "#rotLimit"), 3);
				if rotLimit ~= nil then 
					entry.rotLimit = rotLimit;
				end;
				
				local transLimit = Utils.getVectorNFromString(getXMLString(xmlFile, key .. "#transLimit"), 3);
				if transLimit ~= nil then 
					entry.transLimit = transLimit;
				end;
				
				table.insert(self.parentJoints, entry);
				i = i + 1;
			else
				print("Error: could't find component " .. componentId .. " in '" .. self.configFileName .. "'");
			end;
		end;
	end;
	
	self.oldGame = false;
	self.errorTime = 0;
end;

function customJoints:delete()
	for i in pairs(self.usedJoints) do
		self:deleteJoint(i, true);
	end;
end;

function customJoints:readStream(streamId, connection)
	local numJoints = streamReadInt8(streamId);
	print(g_currentMission.time .. " - customJoints:readStream - numJoints " .. tostring(numJoints));
	
	for i = 1, numJoints do
		local parentJoint = streamReadInt8(streamId);
		local childJoint = streamReadInt8(streamId);
		
		self:makeJoint(parentJoint, childJoint, true, "ReadStream");
		
		print(g_currentMission.time .. " - customJoints:readStream - parent " .. tostring(parentJoint) .. " child " .. tostring(childJoint));
	end;
end;

function customJoints:writeStream(streamId, connection)
	streamWriteInt8(streamId, table.getn(self.usedJoints));
	print(g_currentMission.time .. " - customJoints:writeStream - numJoints " .. tostring(table.getn(self.usedJoints)));
	
	for i = 1, table.getn(self.usedJoints) do
		local joint = self.usedJoints[i];
		
		streamWriteInt8(streamId, joint.parentJoint);
		streamWriteInt8(streamId, joint.childJoint);
		
		print(g_currentMission.time .. " - customJoints:writeStream - parent " .. tostring(joint.parentJoint) .. " child " .. tostring(joint.childJoint));
	end;
end;

function customJoints:mouseEvent(posX, posY, isDown, isUp, button)
end;

function customJoints:keyEvent(unicode, sym, modifier, isDown)
end;

function customJoints:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)
	if not resetVehicles then
		local numCustomJoints = getXMLInt(xmlFile, key .. "#numCustomJoints");
		if numCustomJoints ~= nil and numCustomJoints > 0 then
			for i = 1, numCustomJoints do
				local jointKey = key .. string.format(".customJoint(%d)", i - 1);
				local joint = self.usedJoints[i];
				
				local parent = getXMLInt(xmlFile, jointKey .. "#parent");
				local child = getXMLInt(xmlFile, jointKey .. "#child");
				if parent ~= nil and child ~= nil then
					self:makeJoint(parent, child, false, "Load");
				end;
			end;
		end;
		
		self.oldGame = true;
	end; 

    return BaseMission.VEHICLE_LOAD_OK;
end;

function customJoints:getSaveAttributesAndNodes(nodeIdent)	
	local attributes = "";
	local nodes = "";
	
	if self.usedJoints ~= nil and table.getn(self.usedJoints) > 0 then
		attributes = 'numCustomJoints="' .. table.getn(self.usedJoints) .. '"';
		
		for i = 1, table.getn(self.usedJoints) do
			local joint = self.usedJoints[i];
			
			if joint.jointIndex ~= nil then
				if i > 1 then
					nodes = nodes .. "\n";
				end
				nodes = nodes .. nodeIdent .. '<customJoint';
				nodes = nodes .. ' parent="' .. joint.parentJoint .. '"';
				nodes = nodes .. ' child="' .. joint.childJoint .. '"';
				nodes = nodes .. " />"
			end;
		end;
	end;
	
    return attributes, nodes;
end;

function customJoints:update(dt)
	if not self.oldGame then
		for i = 1, table.getn(self.parentJoints) do
			local p = self.parentJoints[i];
			
			if p.attachChildIdOnLoad ~= nil then
				self:makeJoint(i, p.attachChildIdOnLoad, false, "Load old game");
				p.attachChildIdOnLoad = nil;
			end;
		end;
		
		self.oldGame = true;
	end;
	
	for pI, p in pairs(self.parentJoints) do
		if self.oldGame and p.attachChildIdOnLoad ~= nil then
			p.attachChildIdOnLoad = nil;
		end;
		
		local playerInRange = g_currentMission.player ~= nil and self:jointInRange(p.node, g_currentMission.player.rootNode, 2);
		if (self:getIsActive() or playerInRange) and self.isClient then
			local allowed = true;
			
			for _, a in pairs(p.allow) do
				if not a then
					allowed = false;
					break;
				end;
			end;
			
			for cI, c in pairs(self.childJoints) do
				if p.type == c.type then
					local bool = self:jointInRange(p.node, c.node);
					if bool then
						if (self:getIsActiveForInput(false) and not self:hasInputConflictWithSelection() or playerInRange) then
							local jointIsSet = (self.usedJoints[pI] == nil or self.usedJoints[pI] ~= nil and self.usedJoints[pI].jointIndex == nil);
							
							if InputBinding.hasEvent(InputBinding.ATTACH_JOINT) then
								if allowed then
									if jointIsSet then
										self:makeJoint(pI, cI, false, "Inputbinding");
										self:playAttachSound();
									else
										self:deleteJoint(pI);
										self:playDetachSound();
									end;
								else
									self.errorTime = self.time + 2000;
								end;
							end;
							
							if jointIsSet and allowed then
								g_currentMission:addHelpButtonText(g_i18n:getText("MAKE_JOINT") .. " " .. c.name, InputBinding.ATTACH_JOINT);
							end;
						end;
					end;
				end;
			end;
			
			if self.errorTime > self.time then
				if p.error ~= nil then
					g_currentMission:addWarning(p.error, 0.05, 0.025 + 0.007);
				end;
			else
				if p.error ~= nil then
					p.error = nil;
				end;
			end;
		end;
	end;
end;

function customJoints:updateTick(dt)
end;

function customJoints:draw()
end;

function customJoints:makeJoint(jointId, jointId2, noEventSend, caller)
	print("customJoints:makeJoint - caller " .. tostring(caller));
	
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			-- g_server:broadcastEvent(customJointsMakeEvent:new(self, jointId, jointId2), nil, nil, self);
		else
			-- g_client:getServerConnection():sendEvent(customJointsMakeEvent:new(self, jointId, jointId2));
		end;
	end;
	
	local joint = self.parentJoints[jointId];
	
	if joint ~= nil then
		local usedJoint = {};
		usedJoint.parentJoint = jointId;
		usedJoint.childJoint = jointId2;
		usedJoint.jointNode = joint.node;
		
		if self.isServer then
			local constr = JointConstructor:new();
			constr:setActors(self.components[joint.componentId].node, self.components[self.childJoints[jointId2].componentId].node);
			-- constr:setJointTransforms(joint.node, self.childJoints[jointId2].node);
			constr:setJointTransforms(joint.str, joint.str);
			
			for i = 1, 3 do
				if joint.rotLimit ~= nil then
					constr:setRotationLimit(i-1, Utils.degToRad(-joint.rotLimit[i]), Utils.degToRad(joint.rotLimit[i]));
				else
					constr:setRotationLimit(i-1, 0, 0);
				end;
				
				if joint.transLimit ~= nil then
					constr:setTranslationLimit(i-1, true, -joint.transLimit[i], joint.transLimit[i]);
				else
					constr:setTranslationLimit(i-1, true, 0, 0);
				end;
				
				constr:setLinearDrive(i-1, false, true, 0, 0, 100, 0, 0)
			end;
			constr:setEnableCollision(true)
			
			usedJoint.jointIndex = constr:finalize();
			print("server: joint have been finalize!");
		else
			print("faking joint for client...");
			usedJoint.jointIndex = -1; -- lets give clients a value too.
		end;
		print("storing joint..");
		self.usedJoints[jointId] = usedJoint;
		
		print("calling customJointAttach function");
		for _, v in ipairs(self.specializations) do
			if v.customJointAttach ~= nil then
				v.customJointAttach(self, jointId);
			end;
		end;
		
		print("end!");
		print("");
	else
		print("Error: invalid parentId!");
	end;
end;

function customJoints:deleteJoint(jointId, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			-- g_server:broadcastEvent(customJointsDeleteEvent:new(self, jointId), nil, nil, self);
		else
			-- g_client:getServerConnection():sendEvent(customJointsDeleteEvent:new(self, jointId));
			return;
		end;
	end;
	
	if self.isServer then
		removeJoint(self.usedJoints[jointId].jointIndex);
		print("server deletejoint");
		print("");
	else
		print("client deletejoint");
		print("");
	end;
	self.usedJoints[jointId] = nil;
	
	if self.usedJoints[jointId] ~= nil then
		print("Error: failed to delete JointIndex. " .. jointId)
	end;
	
	for _, v in ipairs(self.specializations) do
		if v.customJointDetach ~= nil then
			v.customJointDetach(self, jointId);
		end;
	end;
end;

function customJoints:jointInRange(object, object2, nd)
	if nd == nil then nd = 0.5; end;
	
	local x1, y1, z1 = getWorldTranslation(object);
	local x2, y2, z2 = getWorldTranslation(object2);
	local distance = Utils.vector3Length(x1 - x2, y1 - y2, z1 - z2);
	
	if distance > nd then
		return false;
	else
		return true;
	end;
end;

function customJoints:updateJoint(id)
	local v = self.usedJoints[id];
	if v ~=  nil and v.jointIndex ~= nil and v.jointNode ~= nil then
		if self.isServer then
			setJointFrame(v.jointIndex, 0, v.jointNode);
		end;
	end;
end;


customJointsMakeEvent = {};
customJointsMakeEvent_mt = Class(customJointsMakeEvent, Event);

InitEventClass(customJointsMakeEvent, "customJointsMakeEvent");

function customJointsMakeEvent:emptyNew()
    print("customJointsMakeEvent:emptyNew");
	
	local self = Event:new(customJointsMakeEvent_mt);
    self.className = "customJointsMakeEvent";
	
    return self;
end;

function customJointsMakeEvent:new(vehicle, jointId, jointId2)
    print("customJointsMakeEvent:new");
	
    local self = customJointsMakeEvent:emptyNew()
    self.vehicle = vehicle;
	self.jointId = jointId;
	self.jointId2 = jointId2;
	
    return self;
end;

function customJointsMakeEvent:readStream(streamId, connection)
    print("customJointsMakeEvent:readStream");
	
    self.vehicle = networkGetObject(streamReadInt32(streamId));
	self.jointId = streamReadInt8(streamId);
	self.jointId2 = streamReadInt8(streamId);
	
	if self.vehicle ~= nil then
		self:run(connection);
	end;
end;

function customJointsMakeEvent:writeStream(streamId, connection)
    print("customJointsMakeEvent:writeStream - obj " .. tostring(networkGetObjectId(self.vehicle)));
	
	streamWriteInt32(streamId, networkGetObjectId(self.vehicle));
	streamWriteInt8(streamId, self.jointId);
	streamWriteInt8(streamId, self.jointId2);
end;

function customJointsMakeEvent:run(connection)
	print("customJointsMakeEvent:run - object " .. tostring(self.vehicle) .. ", " .. tostring(self.jointId) .. ", " .. tostring(self.jointId2));
	
	self.vehicle:makeJoint(self.jointId, self.jointId2, true, "Event");
	
    if not connection:getIsServer() then
		g_server:broadcastEvent(self, nil, connection, self.vehicle);
    end;
end;

customJointsDeleteEvent = {};
customJointsDeleteEvent_mt = Class(customJointsDeleteEvent, Event);

InitEventClass(customJointsDeleteEvent, "customJointsDeleteEvent");

function customJointsDeleteEvent:emptyNew()
    local self = Event:new(customJointsDeleteEvent_mt);
    self.className = "customJointsDeleteEvent";
	
    return self;
end;

function customJointsDeleteEvent:new(vehicle, jointId)
    local self = customJointsDeleteEvent:emptyNew()
    self.vehicle = vehicle;
	self.jointId = jointId;
	
    return self;
end;

function customJointsDeleteEvent:readStream(streamId, connection)
    self.vehicle = networkGetObject(streamReadInt32(streamId));
	self.jointId = streamReadInt8(streamId);
	
    if connection:getIsServer() then
        self.vehicle:deleteJoint(self.jointId, true);
    else
        self.vehicle:deleteJoint(self.jointId);
    end;
	
	print("customJointsDeleteEvent:readStream - object " .. tostring(self.vehicle) .. ", id1 " .. tostring(self.jointId) .. ", " .. tostring(self.jointId2));
end;

function customJointsDeleteEvent:writeStream(streamId, connection)
    print("customJointsDeleteEvent:writeStream - obj " .. tostring(networkGetObjectId(self.vehicle)));
	
    streamWriteInt32(streamId, networkGetObjectId(self.vehicle));
	streamWriteInt8(streamId, self.jointId);
end;