--
-- cylinderInteraction
--
-- @author:    	Xentro (www.fs-uk.com)(Marcus@Xentro.se)
-- @version:    v0.9
-- @date:       2013-07-07
-- 
--[[
<cylinderInteractions> 
	<location trigger="" movingToolId="" />
</cylinderInteractions> 
]]--

cylinderInteraction = {};

function cylinderInteraction.prerequisitesPresent(specializations)
    return true;
end;

function cylinderInteraction:load(xmlFile)
	self.triggerCallback = cylinderInteraction.triggerCallback;
	
	self.interactionLocations = {};
	local i = 0;
	while true do
		local key = string.format("vehicle.cylinderInteractions.location(%d)", i);
		if not hasXMLProperty(xmlFile, key) then break; end;
		
		local triggerNode = Utils.indexToObject(self.components, getXMLString(xmlFile, key .. "#trigger"));
		if triggerNode ~= nil then
			local toolIds = Utils.getVectorNFromString(getXMLString(xmlFile, key .. "#movingToolId"));
			
			if toolIds ~= nil then
				local movingToolIds = {};
				for i = 1, table.getn(toolIds) do
					if self.movingTools[i] ~= nil then
						table.insert(movingToolIds, i);
					end;
				end;
				
				if table.getn(movingToolIds) > 0 then
					local entry = {};
					entry.triggerNode = triggerNode;
					entry.movingToolIds = movingToolIds;
					entry.playerCanControl = false;
					
					addTrigger(entry.triggerNode, "triggerCallback", self);
					
					self.interactionLocations[triggerNode] = entry;
					i = i + 1;
				else
					break;
				end;
			else
				print("Error: cylinderInteractions.location has invalid movingToolId in " .. self.configFileName);
				break;
			end;
		else
			print("Error: cylinderInteractions.location has invalid trigger in " .. self.configFileName);
			break;
		end;
	end;
end;

function cylinderInteraction:delete()
	for _, v in pairs(self.interactionLocations) do
		if v.triggerNode ~= nil then
			removeTrigger(v.triggerNode);
		end;
	end;
end;

function cylinderInteraction:readStream(streamId, connection)
end;

function cylinderInteraction:writeStream(streamId, connection)
end;

function cylinderInteraction:mouseEvent(posX, posY, isDown, isUp, button)
end;

function cylinderInteraction:keyEvent(unicode, sym, modifier, isDown)
end;

function cylinderInteraction:update(dt)	
	if g_gui.currentGui == nil and not g_currentMission.isPlayerFrozen then
		if g_currentMission.controlPlayer and self.isClient then
			for _, v in pairs(self.interactionLocations) do
				if v.playerCanControl then
					for i in ipairs(v.movingToolIds) do
						local tool = self.movingTools[i];
						if tool.isActive and (tool.rotSpeed ~= nil or tool.transSpeed ~= nil) and tool.axisActionIndex ~= nil then
							local move, axisType = InputBinding.getInputAxis(tool.axisActionIndex);
							
							if axisType == InputBinding.INPUTTYPE_MOUSE_AXIS then
								if tool.invertMouseAxis then
									move = -move;
								end;
								
								move = move * tool.speedFactor;
							else
								if tool.invertAxis then
									move = -move;
								end;
							end;
							
							local rotSpeed = 0;
							local transSpeed = 0;
							if not InputBinding.isAxisZero(move) then
								if tool.rotSpeed ~= nil then
									rotSpeed = move * tool.rotSpeed;
									
									if tool.rotAcceleration ~= nil and math.abs(rotSpeed - tool.lastRotSpeed) >= tool.rotAcceleration * dt then
										if rotSpeed > tool.lastRotSpeed then
											rotSpeed = tool.lastRotSpeed + tool.rotAcceleration * dt;
										else
											rotSpeed = tool.lastRotSpeed - tool.rotAcceleration * dt;
										end;
									end;
								end;
								
								if tool.transSpeed ~= nil then
									transSpeed = move * tool.transSpeed;
									
									if tool.transAcceleration ~= nil and math.abs(transSpeed - tool.lastTransSpeed) >= tool.transAcceleration * dt then
										if transSpeed > tool.lastTransSpeed then
											transSpeed = tool.lastTransSpeed + tool.transAcceleration * dt;
										else
											transSpeed = tool.lastTransSpeed - tool.transAcceleration * dt;
										end;
									end;
								end;
							else
								-- decelerate
								if tool.externalRotSpeed ~= 0 then
									rotSpeed = tool.externalRotSpeed;
								elseif tool.rotAcceleration ~= nil then
									if tool.lastRotSpeed < 0 then
										rotSpeed = math.min(tool.lastRotSpeed + tool.rotAcceleration * dt, 0);
									else
										rotSpeed = math.max(tool.lastRotSpeed - tool.rotAcceleration * dt, 0);
									end;
								end;
								
								if tool.externalTransSpeed ~= 0 then
									transSpeed = tool.externalTransSpeed;
								elseif tool.transAcceleration ~= nil then
									if tool.lastTransSpeed < 0 then
										transSpeed = math.min(tool.lastTransSpeed + tool.transAcceleration * dt, 0);
									else
										transSpeed = math.max(tool.lastTransSpeed - tool.transAcceleration * dt, 0);
									end;
								end;
							end;
	  
							local changed = false;
							if rotSpeed ~= 0 then
								local newRot = tool.curRot[tool.rotationAxis] + rotSpeed * dt;
								
								if tool.rotMax ~= nil then
									newRot = math.min(newRot, tool.rotMax);
								end;
								
								if tool.rotMin ~= nil then
									newRot = math.max(newRot, tool.rotMin);
								end;
								
								local diff = newRot - tool.curRot[tool.rotationAxis];
								tool.lastRotSpeed = diff/dt;
								
								if math.abs(diff) > 0.0001 then
									-- wrap if not limited
									if tool.rotMin == nil and tool.rotMax == nil then
										if newRot > 2 * math.pi then
											newRot = newRot - 2 * math.pi;
										end;
										
										if newRot < 0 then
											newRot = newRot + 2 * math.pi;
										end;
									end;
									
									tool.curRot[tool.rotationAxis] = newRot;
									setRotation(tool.node, unpack(tool.curRot));
									changed = true;
								end;
							else
								tool.lastRotSpeed = 0;
							end;
							
							if transSpeed ~= 0 then
								local newTrans = tool.curTrans[tool.translationAxis] + transSpeed * dt;
								
								if tool.transMax ~= nil then
									newTrans = math.min(newTrans, tool.transMax);
								end;
								
								if tool.transMin ~= nil then
									newTrans = math.max(newTrans, tool.transMin);
								end;
								
								local diff = newTrans - tool.curTrans[tool.translationAxis];
								tool.lastTransSpeed = diff / dt;
								
								if math.abs(diff) > 0.0001 then
									tool.curTrans[tool.translationAxis] = newTrans;
									setTranslation(tool.node, unpack(tool.curTrans));
									changed = true;
								end;
							else
								tool.lastTransSpeed = 0;
							end;
		
							for _, dependentTool in pairs(tool.dependentMovingTools) do
								if dependentTool.rotSpeedScale ~= nil and tool.lastRotSpeed ~= 0 then
									dependentTool.movingTool.externalRotSpeed = dependentTool.rotSpeedScale * tool.lastRotSpeed;
								end
								
								if dependentTool.transSpeedScale ~= nil and tool.lastTransSpeed ~= 0 then
									dependentTool.movingTool.externalTransSpeed = dependentTool.transSpeedScale * tool.lastTransSpeed;
								end;
							end;
		
							if changed then
								Cylindered.setDirty(self, tool);
								self:raiseDirtyFlags(self.cylinderedDirtyFlag);
							end;
						end;
						
						tool.externalRotSpeed = 0;
						tool.externalTransSpeed = 0;
					end;
				end;
			end;
		end;
	end;
end;

function cylinderInteraction:updateTick(dt)
end;

function cylinderInteraction:draw()
end;

function cylinderInteraction:triggerCallback(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
	if onEnter or onLeave then
		if g_currentMission.player ~= nil and otherId == g_currentMission.player.rootNode then
			if onEnter then
				self.interactionLocations[triggerId].playerCanControl = true;
			else
				self.interactionLocations[triggerId].playerCanControl = false;
			end;
		end;
	end;
end;